import math
from copy import deepcopy

from auth0.v3.authentication import GetToken
from auth0.v3.management import Auth0
from django.conf import settings
from cached_property import threaded_cached_property_with_ttl
import logging

from django_auth0_user.settings import AUTH0_RULE_CONFIGS
from django_auth0_user.settings import AUTH0_RULES
from django_auth0_user.settings import AUTH0_DOMAIN
from django_auth0_user.settings import AUTH0_API_URL
from django_auth0_user.settings import AUTH0_MANAGEMENT_API_CLIENT_ID
from django_auth0_user.settings import AUTH0_MANAGEMENT_API_CLIENT_SECRET


# TODO: The logging here should be more consistent.
logger = logging.getLogger(__name__)


AUTH0_MANAGEMENT_API_TOKEN_DEFAULT_EXPIRY = 86400  # 24 hours = 86400 seconds
AUTH0_CACHED_TOKEN_TTL = 3600  # 1 hour = 3600 seconds
# AUTH0_DOMAIN = getattr(settings, 'AUTH0_DOMAIN')

# AUTH0_API_URL = 'https://' + AUTH0_DOMAIN + '/api/v2/'
# domain = AUTH0_DOMAIN


# TODO: handle empty domain, client ID and client secret, in a user friendly fashion.


class TokenCache(object):
    """
    Cache the Auth0 token with a TTL so that we get a new one regularly.

    Needed to prevent invalid token issues in production, since the production server needs to possibly live longer
    the token expiry, and we cannot just hammer the token API for a new token every time it makes a request.
    """

    @threaded_cached_property_with_ttl(ttl=AUTH0_CACHED_TOKEN_TTL)
    def auth0_management_api_token(self):
        # TODO: Convert this to a proper logging statement.
        logger.info('No cached Auth0 management API token was found...')
        get_token = GetToken(AUTH0_DOMAIN)
        token = get_token.client_credentials(
            AUTH0_MANAGEMENT_API_CLIENT_ID,
            AUTH0_MANAGEMENT_API_CLIENT_SECRET,
            AUTH0_API_URL
        )
        management_api_token = token['access_token']
        # TODO: Convert this to a proper logging statement.
        logger.info('Successfully generated a new Auth0 Management API Token, this will be cached.')
        return management_api_token


AUTH0_TOKEN_CACHE = TokenCache()


# TODO: Is this the best name for this function?
def get_auth0():
    return Auth0(AUTH0_DOMAIN, AUTH0_TOKEN_CACHE.auth0_management_api_token)


def get_all_auth0_users():
    """
    Update all the user objects...

    :return:
    """
    auth0 = get_auth0()

    auth0_users = []

    first_query = auth0.users.list()
    query_length = first_query['length']
    query_limit = first_query['limit']
    query_start = first_query['start']
    query_total = first_query['total']

    for auth0_user in first_query['users']:
        auth0_users.append(auth0_user)
    if query_length + query_start <= query_total:
        page_number = 0
        while query_length + query_start <= query_total:
            page_number += 1

            page_query = auth0.users.list(page=page_number)

            for auth0_user in page_query['users']:
                auth0_users.append(auth0_user)

            query_length = page_query['length']
            query_limit = page_query['limit']
            query_start = page_query['start']
            query_total = page_query['total']

    return auth0_users


def get_auth0_user(user_id, auth0=None):
    if auth0 is None:
        auth0 = get_auth0()
    # TODO: This does not appear to retrieve all user related data. So it may need to be improved.
    return auth0.users.get(user_id)


def get_users_from_auth0(auth0_conn: Auth0):
    """
    Get all users from Auth0

    :param auth0_conn: Authenticated Auth0 API client
    :param regex: Regex to filter users (re.match(user.email))
    """

    users_list = auth0_conn.users.list()

    total_users = users_list['total']
    page_size = users_list['length']

    # don't waste the first request
    for u in users_list['users']:
        yield u

    del users_list

    # iterate through subsequent pages
    if page_size > 0 and total_users > 0:
        for page in range(1, int(math.ceil(total_users / page_size))):
            for u in auth0_conn.users.list(page=page)['users']:
                yield u


# TODO: Add a function to get an Auth0 client's details from the Management API


# TODO: Add a function that returns true/false based on if the Auth0 client is configured to be OIDC conformant.
# TODO: Work out the best way to use this without incurring major overhead or security risks.
#  One time deployment check command?
def oidc_conformant(client_id):
    auth0 = get_auth0()
    clients = auth0.clients.all(fields=['name', 'client_id', 'oidc_conformant'])
    for client in clients:
        if client['client_id'] == client_id:
            return client['oidc_conformant']


# TODO: Make this part of the setup / deployment somehow ... >_>
# Set Auth0 Rule Configs:
def set_auth0_rule_config_values():
    """
    Add Auth0 Rule Config Data that we have defined.
    :return:
    """
    auth0 = get_auth0()
    for config_key, config_item in AUTH0_RULE_CONFIGS.items():
        result = auth0.rules_configs.set(config_key, config_item)
        logger.info(f"Auth0 Set Rule Config Key Result: result={result}")


def remove_auth0_rule_config_values():
    """
    Remove any Auth0 Rule Config data that we have setup.

    :return: None
    """
    auth0 = get_auth0()
    for config_key, config_item in AUTH0_RULE_CONFIGS.items():
        auth0.rules_configs.unset(config_key)


# Auth0 Rules:
# TODO: Document and put in some code to help Limit the required permissions when using the management API.
#  This is super important for being able to run a CI suite against this!
#  Even if stored correctly the API keys placed in a CI suite are
#  inherently vulnerable so should be limited as much as possible.

# TODO: If stage doesnt match, the rule will need to be deleted & then added again.
# TODO: It doesnt seem possible to set the stage directly... So Should I ignore it?
def setup_auth0_rules(dry_run=True):
    """
    Setup Auth0 Rules using the management API.

    :param dry_run:
    :return:
    """
    auth0 = get_auth0()

    current_rule_list = auth0.rules.all()
    current_rules = {}
    for rule in current_rule_list:
        rule_name = rule['name']
        _rule = deepcopy(rule)
        del _rule['name']
        current_rules[rule['name']] = {**_rule}
    current_rule_names = set(current_rules.keys())

    settings_rules = AUTH0_RULES
    settings_rule_names = set(settings_rules.keys())

    # If the rule isn't one we have defined, ignore it.
    # rules_to_ignore = current_rule_names - settings_rule_names
    # If the rule is one we defined and it doesnt exist yet, we need to create it.
    rules_to_create = settings_rule_names - current_rule_names
    # If the rule is defined and already exists, we need to check it is configured correctly.
    rules_to_check = current_rule_names & settings_rule_names
    rules_to_update = set()

    for rule_name in rules_to_create:
        if not dry_run:
            auth0.rules.create({
                'name': rule_name,
                **settings_rules[rule_name]
            })

    for rule_name in rules_to_check:
        if settings_rules[rule_name]['enabled'] == current_rules[rule_name]['enabled']:
            if settings_rules[rule_name]['order'] == current_rules[rule_name]['order']:
                if settings_rules[rule_name]['script'] == current_rules[rule_name]['script']:
                    continue
        rules_to_update.add(rule_name)

    # TODO: update rules here.
    for rule_name in rules_to_update:
        if not dry_run:
            auth0.rules.update(rule_name, {
                'name': rule_name,
                **settings_rules[rule_name]
            })


def tear_down_auth0_rules(dry_run=True):
    auth0 = get_auth0()
    # ----------------------------------------
    settings_rules = AUTH0_RULES
    settings_rule_names = set(settings_rules.keys())
    # ----------------------------------------
    current_rule_mapping = {}
    for rule in auth0.rules.all():
        current_rule_mapping[rule['name']] = rule['id']
    for rule_name in settings_rule_names:
        rule_id = current_rule_mapping[rule_name]
        if not dry_run:
            auth0.rules.delete(rule_id)



from django.conf import settings


if getattr(settings, 'AUTH0_DOMAIN', None) is not None:
    AUTH0_DOMAIN = settings.AUTH0_DOMAIN
elif getattr(settings, 'SOCIAL_AUTH_AUTH0_DOMAIN', None) is not None:
    AUTH0_DOMAIN = settings.SOCIAL_AUTH_AUTH0_DOMAIN
else:
    AUTH0_DOMAIN = None

AUTH0_API_URL = 'https://' + AUTH0_DOMAIN + '/api/v2/'

if getattr(settings, 'AUTH0_OIDC_ENDPOINT', None) is not None:
    AUTH0_OIDC_ENDPOINT = settings.AUTH0_OIDC_ENDPOINT
elif getattr(settings, 'SOCIAL_AUTH_AUTH0_OIDC_ENDPOINT', None) is not None:
    AUTH0_OIDC_ENDPOINT = settings.SOCIAL_AUTH_AUTH0_OIDC_ENDPOINT
else:
    AUTH0_OIDC_ENDPOINT = None

if getattr(settings, 'AUTH0_USER_ID_IS_DJANGO_USERNAME', None) is not None:
    USER_ID_IS_DJANGO_USERNAME = settings.AUTH0_USER_ID_IS_DJANGO_USERNAME
if getattr(settings, 'SOCIAL_AUTH_AUTH0_USER_ID_IS_DJANGO_USERNAME', None) is not None:
    USER_ID_IS_DJANGO_USERNAME = settings.SOCIAL_AUTH_AUTH0_USER_ID_IS_DJANGO_USERNAME
else:
    USER_ID_IS_DJANGO_USERNAME = True


#


# ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
# ┃  Auth0 Management API Settings  ┃▓▓
# ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛▓▓
#   ▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓
if getattr(settings, 'AUTH0_MANAGEMENT_API_CLIENT_ID', None) is not None:
    AUTH0_MANAGEMENT_API_CLIENT_ID = settings.AUTH0_MANAGEMENT_API_CLIENT_ID
if getattr(settings, 'SOCIAL_AUTH_AUTH0_MANAGEMENT_API_CLIENT_ID', None) is not None:
    AUTH0_MANAGEMENT_API_CLIENT_ID = settings.SOCIAL_AUTH_AUTH0_MANAGEMENT_API_CLIENT_ID
else:
    AUTH0_MANAGEMENT_API_CLIENT_ID = None
if getattr(settings, 'AUTH0_MANAGEMENT_API_CLIENT_SECRET', None) is not None:
    AUTH0_MANAGEMENT_API_CLIENT_SECRET = settings.AUTH0_MANAGEMENT_API_CLIENT_SECRET
if getattr(settings, 'SOCIAL_AUTH_AUTH0_MANAGEMENT_API_CLIENT_SECRET', None) is not None:
    AUTH0_MANAGEMENT_API_CLIENT_SECRET = settings.SOCIAL_AUTH_AUTH0_MANAGEMENT_API_CLIENT_SECRET
else:
    AUTH0_MANAGEMENT_API_CLIENT_SECRET = None


#


# We need a namespace prefix provided by the user as there is not really a safe and unique default.
if getattr(settings, 'AUTH0_NAMESPACED_KEY_PREFIX', None) is not None:
    NAMESPACED_KEY_PREFIX = settings.AUTH0_NAMESPACED_KEY_PREFIX
else:
    NAMESPACED_KEY_PREFIX = ''


if getattr(settings, 'AUTH0_NAMESPACED_USER_METADATA_KEY', None) is not None:
    NAMESPACED_USER_METADATA_KEY = settings.AUTH0_NAMESPACED_USER_METADATA_KEY
elif getattr(settings, 'SOCIAL_AUTH_AUTH0_NAMESPACED_USER_METADATA_KEY', None) is not None:
    NAMESPACED_USER_METADATA_KEY = settings.SOCIAL_AUTH_AUTH0_NAMESPACED_USER_METADATA_KEY
else:
    NAMESPACED_USER_METADATA_KEY = NAMESPACED_KEY_PREFIX + '/user_metadata'

if getattr(settings, 'AUTH0_NAMESPACED_APP_METADATA_KEY', None) is not None:
    NAMESPACED_APP_METADATA_KEY = settings.AUTH0_NAMESPACED_APP_METADATA_KEY
elif getattr(settings, 'SOCIAL_AUTH_AUTH0_NAMESPACED_APP_METADATA_KEY', None) is not None:
    NAMESPACED_APP_METADATA_KEY = settings.SOCIAL_AUTH_AUTH0_NAMESPACED_APP_METADATA_KEY
else:
    NAMESPACED_APP_METADATA_KEY = NAMESPACED_KEY_PREFIX + '/app_metadata'


DEFAULT_AUTH0_RULE_CONFIGS = {
        'DJANGO_AUTH0_USER_OIDC_NAMESPACE_PREFIX': NAMESPACED_KEY_PREFIX,
        'DJANGO_AUTH0_USER_NAMESPACED_USER_METADATA_KEY': NAMESPACED_USER_METADATA_KEY,
        'DJANGO_AUTH0_USER_NAMESPACED_APP_METADATA_KEY': NAMESPACED_APP_METADATA_KEY,
    }
if getattr(settings, 'AUTH0_RULE_CONFIGS', None) is not None:
    AUTH0_RULE_CONFIGS = settings.AUTH0_RULE_CONFIGS
elif getattr(settings, 'SOCIAL_AUTH_AUTH0_RULE_CONFIGS', None) is not None:
    AUTH0_RULE_CONFIGS = settings.SOCIAL_AUTH_AUTH0_RULE_CONFIGS
else:
    AUTH0_RULE_CONFIGS = DEFAULT_AUTH0_RULE_CONFIGS


# TODO: Decide if this should keep living here or belongs in test settings...
# Use the Auth0 Rule Configuration object to set the namespace prefix dynamically.
EXAMPLE_METADATA_RULE_FUNCTION = """function (user, context, callback) {
    const NAMESPACE_PREFIX = configuration.DJANGO_AUTH0_USER_OIDC_NAMESPACE_PREFIX;
    const NAMESPACED_USER_METADATA_KEY = configuration.DJANGO_AUTH0_USER_NAMESPACED_USER_METADATA_KEY;
    const NAMESPACED_APP_METADATA_KEY = configuration.DJANGO_AUTH0_USER_NAMESPACED_APP_METADATA_KEY;
    if (context.idToken && user.user_metadata) {
        context.idToken[NAMESPACED_USER_METADATA_KEY] = user.user_metadata;
    }
    if (context.idToken && user.app_metadata) {
        context.idToken[NAMESPACED_APP_METADATA_KEY] = user.app_metadata;
    }
    if (context.accessToken && user.user_metadata) {
        context.accessToken[NAMESPACED_USER_METADATA_KEY] = user.user_metadata;
    }
    if (context.accessToken && user.app_metadata) {
        context.accessToken[NAMESPACED_APP_METADATA_KEY] = user.app_metadata;
    }
    callback(null, user, context);
}
"""


DEFAULT_AUTH0_RULES = {
        "Django-Auth0-User-ExampleMetadataRule": {
            "script": EXAMPLE_METADATA_RULE_FUNCTION,
            "stage": "login_success",
            "enabled": True,
            "order": 1,
        }
    }
if getattr(settings, 'AUTH0_RULES', None) is not None:
    AUTH0_RULES = settings.AUTH0_RULES
elif getattr(settings, 'SOCIAL_AUTH_AUTH0_RULES', None) is not None:
    AUTH0_RULES = settings.SOCIAL_AUTH_AUTH0_RULES
else:
    AUTH0_RULES = DEFAULT_AUTH0_RULES



from django.conf import settings
from django.contrib.auth.views import redirect_to_login
from django.contrib.admin import AdminSite
from django.utils.http import is_safe_url
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.shortcuts import resolve_url
from django_auth0_user.util import monkeypatch_method

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode


class AdminSiteAuth0LoginMixin(object):
    """
    Mixin that uses Auth0 for login and Auth0 user data for permissions.
    """

    def login(self, request, extra_context=None):
        """
        Redirects to the site login page for the given HttpRequest.

        :param request:
        :param extra_context:
        :return:
        """
        redirect_to = request.POST.get(REDIRECT_FIELD_NAME, request.GET.get(REDIRECT_FIELD_NAME))

        if not redirect_to or not is_safe_url(url=redirect_to, host=request.get_host()):
            redirect_to = resolve_url(settings.LOGIN_REDIRECT_URL)

        return redirect_to_login(redirect_to)


class AdminSiteWithAuth0Authentication(AdminSiteAuth0LoginMixin, AdminSite):
    """
    AdminSite using Auth0 for authentication and Auth0 user data for permission checking.
    """
    pass


# Should I be using wrapt?
def patch_admin():
    @monkeypatch_method(AdminSite)
    def login(self, request, extra_context=None):
        """
        Redirects to the site login page for the given HttpRequest.

        :param self:
        :param request:
        :param extra_context:
        :return:
        """
        redirect_to = request.POST.get(REDIRECT_FIELD_NAME, request.GET.get(REDIRECT_FIELD_NAME))

        if not redirect_to or not is_safe_url(url=redirect_to, host=request.get_host()):
            redirect_to = resolve_url(settings.LOGIN_REDIRECT_URL)

        return redirect_to_login(redirect_to)

        # return redirect('%s?%s' % (
        #     settings.LOGIN_URL,
        #     urlencode({REDIRECT_FIELD_NAME: request.get_full_path()})
        # ))


def unpatch_admin():
    setattr(AdminSite, 'login', original_login)


original_login = AdminSite.login
if getattr(settings, 'AUTH0_PATCH_ADMIN', True):
    patch_admin()



import importlib
from django.conf import settings


def auth0_user_is_superuser(auth0_user):
    """
    Check if an Auth0User has been marked as having superuser status.

    :param django_db_auth0_user.models.Auth0User auth0_user: The Auth0User to check for supervisor status.
    :return:
    """
    if auth0_user.auth0_app_metadata:
        return auth0_user.auth0_app_metadata.get('is_superuser', False)
    else:
        return False


def auth0_user_is_staff(auth0_user):
    """
    Check if an Auth0User has been marked as having staff status.

    :param django_db_auth0_user.models.Auth0User auth0_user: The Auth0User to check for staff status.
    :return:
    """
    if auth0_user.auth0_app_metadata:
        if auth0_user.is_superuser:
            return True
        return auth0_user.auth0_app_metadata.get('is_staff', False)
    else:
        return False


def auth0_user_is_active(auth0_user):
    """
    Check if an Auth0User has been marked as having active status.

    The built in default is True, since the majority of sites will want to set new users as active.

    :param django_db_auth0_user.models.Auth0User auth0_user: The Auth0User to check for active status.
    :return:
    """
    default_active_value = getattr(settings, 'AUTH0_USER_DEFAULT_IS_ACTIVE_VALUE', True)
    if auth0_user.auth0_app_metadata:
        return auth0_user.auth0_app_metadata.get('is_active', default_active_value)
    else:
        return default_active_value


# Just a reminder how this import magic works.
# function_string = 'mypackage.mymodule.myfunc'
# mod_name, func_name = function_string.rsplit('.',1)
# mod = importlib.import_module(mod_name)
# func = getattr(mod, func_name)
# result = func()


AUTH0_USER_IS_SUPERUSER_CHECK_FUNCTION = getattr(settings, 'AUTH0_USER_IS_SUPERUSER_CHECK_FUNCTION',
                                                 'django_auth0_user.permission_checks.auth0_user_is_superuser')
superuser_module_name, supervisor_func_name = AUTH0_USER_IS_SUPERUSER_CHECK_FUNCTION.rsplit('.', 1)
superuser_module = importlib.import_module(superuser_module_name)
IS_SUPERUSER = getattr(superuser_module, supervisor_func_name)


AUTH0_USER_IS_STAFF_CHECK_FUNCTION = getattr(settings, 'AUTH0_USER_IS_STAFF_CHECK_FUNCTION',
                                             'django_auth0_user.permission_checks.auth0_user_is_staff')
staff_module_name, staff_func_name = AUTH0_USER_IS_STAFF_CHECK_FUNCTION.rsplit('.', 1)
staff_module = importlib.import_module(staff_module_name)
IS_STAFF = getattr(staff_module, staff_func_name)


AUTH0_USER_IS_ACTIVE_CHECK_FUNCTION = getattr(settings, 'AUTH0_USER_IS_ACTIVE_CHECK_FUNCTION',
                                              'django_auth0_user.permission_checks.auth0_user_is_active')
active_module_name, active_func_name = AUTH0_USER_IS_ACTIVE_CHECK_FUNCTION.rsplit('.', 1)
active_module = importlib.import_module(active_module_name)
IS_ACTIVE = getattr(active_module, active_func_name)











import six
import unicodedata
import re


# This slugify version was borrowed from django revision a61dbd6
def slugify(value):
    """Converts to lowercase, removes non-word characters (alphanumerics
    and underscores) and converts spaces to hyphens. Also strips leading
    and trailing whitespace."""
    value = unicodedata.normalize('NFKD', six.text_type(value)).encode('ascii', 'ignore').decode('ascii')
    value = re.sub('[^[\w.@+-|]', '', value).strip().lower()
    return re.sub('[-\s]+', '-', value)



from django.core.management.base import BaseCommand, CommandError
from django_auth0_user.util.auth0_api import set_auth0_rule_config_values


# TODO: It should also be possible to set the Auth0 Management API Credentials
#  here instead of requiring them in settings.
class Command(BaseCommand):
    help = 'Add the values specified in AUTH0_RULE_CONFIGS to the Auth0 Rule Configs using the Auth0 Management API'

    def handle(self, *args, **options):
        set_auth0_rule_config_values()



from django.core.management.base import BaseCommand, CommandError
from django_auth0_user.util.auth0_api import tear_down_auth0_rules


# TODO: It should also be possible to set the Auth0 Management API Credentials
#  here instead of requiring them in settings.
class Command(BaseCommand):
    help = 'Remove the values specified in AUTH0_RULES from the Rules in Auth0 using the Auth0 Management API'

    def handle(self, *args, **options):
        tear_down_auth0_rules(dry_run=False)



# -*- coding: utf-8 -*-
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from rest_framework.authentication import BaseAuthentication
from rest_framework.authentication import get_authorization_header
from rest_framework import exceptions
from rest_framework import HTTP_HEADER_ENCODING
from social_core.exceptions import MissingBackend
from social_core.exceptions import AuthForbidden
from social_core.utils import requests
from social_django.views import NAMESPACE
from social_django.utils import load_backend
from social_django.utils import load_strategy
import jwt
from rest_framework_jwt.authentication import JSONWebTokenAuthentication
from rest_framework_jwt.authentication import get_user_model
from rest_framework_jwt.authentication import jwt_get_username_from_payload
from rest_framework_jwt.authentication import jwt_decode_handler


class FastAuth0Authentication(JSONWebTokenAuthentication):
    """
    A fast django rest framework authentication class powered by djangorestframework-jwt.

    This only performs JWT validation and user lookup, it will not work for new
    users which is why it needs to be used along with the full drf auth class
    and must be placed before the full class in the DRF auth classes list.
    Clients authenticate using the token key in the "Authorization" HTTP header. For example:

        Authorization: Bearer g87ewgbgg78wegpngxgfoawgbo78aegnafxngfe
    """
    def authenticate_credentials(self, payload):
        """
        Returns an active user that matches the payload's user id and email.
        """
        User = get_user_model()
        username = jwt_get_username_from_payload(payload)

        if not username:
            msg = _('Invalid payload.')
            raise exceptions.AuthenticationFailed(msg)

        try:
            user = User.objects.get_by_natural_key(username)
        except User.DoesNotExist:
            return  # Fallthru

        if not user.is_active:
            msg = _('User account is disabled.')
            raise exceptions.AuthenticationFailed(msg)

        return user

    def authenticate(self, request):
        """
        Returns a two-tuple of `User` and token if a valid signature has been
        supplied using JWT-based authentication.  Otherwise returns `None`.
        """
        jwt_value = self.get_jwt_value(request)
        if jwt_value is None:
            return None

        try:
            payload = jwt_decode_handler(jwt_value)
        except jwt.ExpiredSignature:
            msg = _('Signature has expired.')
            raise exceptions.AuthenticationFailed(msg)
        except jwt.DecodeError:
            msg = _('Error decoding signature.')
            raise exceptions.AuthenticationFailed(msg)
        except jwt.InvalidTokenError:
            raise exceptions.AuthenticationFailed()

        user = self.authenticate_credentials(payload)

        if not user:
            return  # Fall-Through to next auth system

        return user, jwt_value


class FullAuth0Authentication(BaseAuthentication):
    """
    A full django rest framework authentication class powered by `python-social-auth`.

    This completes the full authentication process using PSA. This includes
    fetching profile details etc, from Auth0, as a consequence this can take
    a second or two to complete the full round trip.
    Clients authenticate using the token key in the "Authorization" HTTP header. For example:

        Authorization: Bearer g87ewgbgg78wegpngxgfoawgbo78aegnafxngfe
    """
    # TODO: This should probably be configured via a setting key.
    www_authenticate_realm = 'api'

    def authenticate(self, request):
        """
        Authenticate the request and return a two-tuple of (user, token).
        """
        auth_header = get_authorization_header(request).decode(HTTP_HEADER_ENCODING)

        auth_parts = auth_header.split()

        if len(auth_parts) == 0:
            # No authentication header means we have nothing to work with.
            return None
        elif len(auth_parts) == 1:
            msg = 'Invalid token header. Credentials not provided.'
            raise exceptions.AuthenticationFailed(msg)
        elif len(auth_parts) > 2:
            msg = 'Invalid token header. The token string cannot contain spaces.'
            raise exceptions.AuthenticationFailed(msg)

        auth_type, auth_token = auth_parts[0], auth_parts[1]

        if auth_type.lower() != 'bearer':
            msg = 'Invalid authentication header type. Only Bearer tokens are currently supported..'
            raise exceptions.AuthenticationFailed(msg)

        strategy = load_strategy(request=request)

        try:
            backend = load_backend(strategy, 'auth0', reverse(NAMESPACE + ":complete", args=('auth0',)))
        except MissingBackend:
            msg = 'Either token header is invalid or the backend could not be loaded.'
            raise exceptions.AuthenticationFailed(msg)

        try:
            user = backend.do_auth(access_token=auth_token)
        except AuthForbidden as err:
            raise exceptions.AuthenticationFailed(err)
        except requests.HTTPError as e:
            msg = e.response.text
            raise exceptions.AuthenticationFailed(msg)

        if not user:
            msg = 'Unable to authenticate these credentials.'
            raise exceptions.AuthenticationFailed(msg)
        return user, auth_token

    def authenticate_header(self, request):
        """
        Return a string to be used as the value of the `WWW-Authenticate`
        header in a `401 Unauthenticated` response, or `None` if the
        authentication scheme should return `403 Permission Denied` responses.
        """
        return 'Bearer realm="%s"' % self.www_authenticate_realm



from __future__ import unicode_literals

from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import ugettext_lazy as _

from jwt import decode as jwt_decode
from jwt import DecodeError

from django_auth0_user.validators import Auth0UserIdValidator
from django_auth0_user.settings import USER_ID_IS_DJANGO_USERNAME
from django_auth0_user.settings import NAMESPACED_USER_METADATA_KEY
from django_auth0_user.settings import NAMESPACED_APP_METADATA_KEY


# TODO: The default social_django model uses a foreign key from the social account model
#   to a django user model instance in order to allow multiple social logins for a single
#   django user, however since this is a function that Auth0 abstracts away for us,
#   we can bind the models more tightly and use a OneToOneField. Look into this option further.


class AbstractAuth0User(AbstractUser):
    """
    An abstract base user designed for easy use with Auth0
    """

    # In order to accept the "|" character in a Django Username we need to change the validator.
    if USER_ID_IS_DJANGO_USERNAME:
        username_validator = Auth0UserIdValidator()
        username_help_text = _('Required. 150 characters or fewer. Letters, digits and @/./+/-/_/| only.')
    else:
        username_validator = UnicodeUsernameValidator()
        username_help_text = _('Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.')

    username = models.CharField(
        _('username'),
        max_length=150,
        unique=True,
        help_text=username_help_text,
        validators=[username_validator],
        error_messages={
            'unique': _("A user with that username already exists."),
        },
    )

    # ----------------
    # Helper Functions
    # ----------------

    @property
    def auth0_data(self):
        if self.social_auth.count() == 1:
            return self.social_auth.get().extra_data
        else:
            raise NotImplementedError(
                "More than one social auth model instance is associated with this django user model instance"
            )

    @property
    def id_token_payload(self):
        if self.auth0_data['id_token'] is not None:
            try:
                # TODO: Decide how I want to handle verifying the tokens.
                return jwt_decode(self.auth0_data['id_token'], verify=False)
            except DecodeError:
                return None
        else:
            return None

    @property
    def access_token_payload(self):
        if self.auth0_data['access_token'] is not None:
            try:
                # TODO: Decide how I want to handle verifying the tokens.
                return jwt_decode(self.auth0_data['access_token'], verify=False)
            except DecodeError:
                return None
        else:
            return None

    @property
    def refresh_token_payload(self):
        if self.auth0_data['refresh_token'] is not None:
            try:
                # TODO: Decide how I want to handle verifying the tokens.
                return jwt_decode(self.auth0_data['refresh_token'], verify=False)
            except DecodeError:
                return None
        else:
            return None

    # -------------------------------------
    # Convenience / Quick Access Properties
    # -------------------------------------
    # Accessing user data to make decisions should be quick and easy,
    # so our abstract class adds these properties to help make it easier
    # to leverage Auth0 functionality such as the user and app metadata.

    @property
    def user_metadata(self):
        # TODO: Only do this is we are dealing with an OIDC compliant endpoint.
        # TODO: Ensure any auto-created rule is based on the same metadata dict key so this doesnt break.

        if NAMESPACED_USER_METADATA_KEY is not None:
            if NAMESPACED_USER_METADATA_KEY in self.auth0_data['id_token_payload']:
                return self.auth0_data['id_token_payload'][NAMESPACED_USER_METADATA_KEY]

        if 'user_metadata' in self.auth0_data['id_token_payload']:
            return self.auth0_data['id_token_payload']['user_metadata']

        return None

    @property
    def app_metadata(self):
        # TODO: Only do this is we are dealing with an OIDC compliant endpoint.
        # TODO: Ensure any auto-created rule is based on the same metadata dict key so this doesnt break.

        if NAMESPACED_APP_METADATA_KEY is not None:
            if NAMESPACED_APP_METADATA_KEY in self.auth0_data['id_token_payload']:
                return self.auth0_data['id_token_payload'][NAMESPACED_APP_METADATA_KEY]

        if 'app_metadata' in self.auth0_data['id_token_payload']:
            return self.auth0_data['id_token_payload']['app_metadata']

        return None

    class Meta:
        abstract = True


# TODO: Move this into tests, or raise a warning when used?
# Users of this library should have their own custom user model!



# from http://mail.python.org/pipermail/python-dev/2008-January/076194.html
def monkeypatch_method(cls):
    def decorator(func):
        setattr(cls, func.__name__, func)
        return func
    return decorator



from social_core.backends.open_id_connect import OpenIdConnectAuth
import logging
from six.moves.urllib_parse import urlencode, unquote
from django_auth0_user.settings import AUTH0_OIDC_ENDPOINT

from jose import jwk, jwt
from jose.utils import base64url_decode

# from django.conf import settings
# from social_core.exceptions import AuthCanceled, AuthTokenError, AuthFailed
# import base64, json, hmac, hashlib
# from calendar import timegm
# from datetime import datetime

log = logging.getLogger(__name__)


# TODO: Handle both OpenID Compliant and non-OpenID compliant Auth0 clients!
# ^ This may need some clever test setup to test against 2 sets of Auth0 client details.

# TODO: Add an option to ensure that metadata is returned via rules, using the Management API to set them up for users.
# https://auth0.com/docs/api/management/v2#!/Rules/get_rules

# TODO: Use Django's checks framework/tools to warn about misconfiguration.
#       e.g. OIDC Compliance + No Rule to include Metadata + Some future feature that depends on metadata.


# Valid Lock v10 Configuration Parameters via Query String Arguments
# http://auth0.com/docs/libraries/lock/v10/configuration
AUTH0_ALLOWED_URL_PARAMETERS = (
    # Display
    'allow_autocomplete',
    'allowed_connections',
    'allow_show_password',
    'autoclose',
    'autofocus',
    'avatar',
    'closable',
    'container',
    'language',
    'language_dictionary',
    'popup_options',
    'remember_last_login',

    # Theming
    'theme',
    'auth_buttons',
    'labeled_submit_button',
    'logo',
    'primary_color',

    # Social
    'social_button_style',

    # Authentication
    'auth'
    'audience',
    'auto_parse_hash',
    'connection_scopes',
    'params',
    'redirect_url',
    'response_mode',
    'response_type',
    'sso',

    # Database
    'additional_sign_up_fields',
    'allow_login',
    'allow_forgot_password',
    'allow_sign_up',
    'default_database_connection',
    'initial_screen',
    'login_after_sign_up',
    'forgot_password_link',
    'must_accept_terms',
    'prefill',
    'sign_up_link',
    'username_style',

    # Enterprise
    'default_enterprise_connection',

    # Other
    'oidc_conformant',
    'client_base_url',
    'language_base_url',
    'hash_cleanup',
    'leeway',
)


class Auth0OpenId(OpenIdConnectAuth):
    """Auth0 OpenID authentication backend"""
    name = 'auth0'

    OIDC_ENDPOINT = AUTH0_OIDC_ENDPOINT
    ID_TOKEN_ISS = AUTH0_OIDC_ENDPOINT + "/"
    # Under some circumstances we wont have the 'user_id' in the response during get_user_details.
    # One option is to re-implement this function here in our backend
    # for completeness and ease of documenting how this all works...
    # The other is setting USERNAME_KEY = 'sub' instead of the more obvious USERNAME_KEY = 'user_id',
    # were going with the other for now to keep things simple.
    USERNAME_KEY = 'sub'

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        """Return access_token, token_type, and extra defined names to store in
            extra_data field"""
        data = super(Auth0OpenId, self).extra_data(user, uid, response, details=details, *args, **kwargs)
        # TODO work out why the id_token here has not been decoded
        #  because there may be a better place to do this...
        #  https://github.com/python-social-auth/social-core/issues/127
        # TODO: Don't just store everything here, be more selective.
        data['id_token_payload'] = self.id_token
        return data

    def auth_url(self):
        """Return redirect url"""
        state = self.get_or_create_state()
        params = self.auth_params(state)
        params.update(self.get_scope_argument())
        params.update(self.auth_extra_arguments())
        params.update({_key: self.data[_key] for _key in self.data.keys() if _key in AUTH0_ALLOWED_URL_PARAMETERS})
        params = urlencode(params)
        if not self.REDIRECT_STATE:
            # redirect_uri matching is strictly enforced, so match the
            # providers value exactly.
            params = unquote(params)
        return '{0}?{1}'.format(self.authorization_url(), params)



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': 'src'}

packages = \
['django_auth0_user',
 'django_auth0_user.management',
 'django_auth0_user.management.commands',
 'django_auth0_user.migrations',
 'django_auth0_user.rest_framework',
 'django_auth0_user.templatetags',
 'django_auth0_user.util']

package_data = \
{'': ['*']}

install_requires = \
['cached-property>=1.5.1,<2.0.0',
 'social-auth-app-django>=3.1.0,<4.0.0',
 'social-auth-core[openidconnect]>=3.2.0,<4.0.0']

extras_require = \
{'drf': ['djangorestframework>=3.10.3,<4.0.0',
         'djangorestframework-jwt>=1.11.0,<2.0.0',
         'pyjwt>=1.7.1,<2.0.0']}

setup_kwargs = {
    'name': 'django-auth0-user',
    'version': '0.15.5',
    'description': 'Django Authentication and Authorisation using Auth0 and Python Social Auth',
    'long_description': None,
    'author': 'Samuel Bishop',
    'author_email': 'sam@techdragon.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/techdragon/django-auth0-user',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
